import { HttpException, HttpStatus, Injectable, Logger } from '@nestjs/common';
import { InjectModel } from '@nestjs/mongoose';
import { Cron, CronExpression } from '@nestjs/schedule';
import { Model } from 'mongoose';
import { IMailchimpSubscription } from './interface/mailchimp-subscription';
import { INewsletterSubscription } from './interface/newsletter-subscription.interface';
import { NewsletterSubscription, NewsletterSubscriptionDocument } from './newsletter-subscription.schema';
// eslint-disable-next-line @typescript-eslint/no-var-requires
const mailchimp = require('@mailchimp/mailchimp_marketing');
@Injectable()
export class NewsletterService {
  private readonly logger = new Logger(NewsletterService.name);
  private LIST_ID = process.env.MC_LIST_ID;
  constructor(
    @InjectModel(NewsletterSubscription.name) private newsletterSubscriptionModel: Model<INewsletterSubscription>
  ) {
    // Configure mailchimp client
    mailchimp.setConfig({
      apiKey: process.env.MC_API_KEY,
      server: process.env.MC_SERVER,
    });
  }

  @Cron(CronExpression.EVERY_DAY_AT_3AM)
  public async updateNewsletterSubscription(): Promise<void> {
    this.logger.debug('updateNewsletterSubscription');
    const { total_items } = await mailchimp.lists.getListMembersInfo(this.LIST_ID);
    const { members } = await mailchimp.lists.getListMembersInfo(this.LIST_ID, {
      fields: ['members.email_address,members.id,members.status'],
      count: total_items,
    });
    const memberToRemove = members.filter((user: IMailchimpSubscription) => user.status !== 'subscribed');

    memberToRemove.forEach(async (member: IMailchimpSubscription) => {
      const userSubscription = await this.findOne(member.email_address);
      if (userSubscription) {
        this.logger.log(`Remove subscription ${member.id}`);
        userSubscription.deleteOne();
      }
    });
  }

  public async newsletterSubscribe(email: string): Promise<NewsletterSubscription> {
    this.logger.debug('newsletterSubscribe');
    const existingEmail = await this.findOne(email);
    if (existingEmail) {
      throw new HttpException('Email already exists', HttpStatus.BAD_REQUEST);
    }
    try {
      const member = await mailchimp.lists.addListMember(this.LIST_ID, {
        email_address: email,
        status: 'subscribed',
      });
      await this.newsletterSubscriptionModel.create({ email: email, mailchimpId: member.id });
      return this.findOne(email);
    } catch (e) {
      if (e.status === 400) {
        this.logger.error(`Error ${e.status}, user might already exist in mailchimplist`);
        throw new HttpException('Email already exists', HttpStatus.BAD_REQUEST);
      } else {
        this.logger.error(`Mailchimp configuration error`);
        throw new HttpException('Server error', HttpStatus.INTERNAL_SERVER_ERROR);
      }
    }
  }

  public async newsletterUnsubscribe(email: string): Promise<NewsletterSubscription> {
    this.logger.debug('newsletterUnsubscribe');
    const subscription = await this.findOne(email);
    if (!subscription) {
      throw new HttpException('Invalid email', HttpStatus.BAD_REQUEST);
    }
    await mailchimp.lists.setListMember(this.LIST_ID, subscription.mailchimpId, {
      email_address: email,
      status: 'unsubscribed',
    });
    return subscription.deleteOne();
  }

  public async findOne(mail: string): Promise<INewsletterSubscription | undefined> {
    this.logger.debug('findOne');
    return this.newsletterSubscriptionModel.findOne({ email: mail });
  }

  public async searchNewsletterSubscription(searchString: string): Promise<NewsletterSubscriptionDocument[]> {
    this.logger.debug('searchNewsletterSubscription');
    return this.newsletterSubscriptionModel.find({ email: new RegExp(searchString, 'i') });
  }

  public async countNewsletterSubscriptions(): Promise<number> {
    this.logger.debug('countNewsletterSubscriptions');
    return this.newsletterSubscriptionModel.countDocuments({});
  }

  public async findAll(): Promise<NewsletterSubscription[]> {
    this.logger.debug('findAll');
    return this.newsletterSubscriptionModel.find();
  }
}
