import { Injectable, Logger } from '@nestjs/common';
import { ElasticsearchService } from '@nestjs/elasticsearch';
import { Structure, StructureDocument } from '../schemas/structure.schema';
import { StructureSearchBody } from '../interfaces/structure-search-body.interface';
import { StructureSearchResult } from '../interfaces/structure-search-response.interface';
import { es_settings_homemade_french } from '../../shared/utils';

@Injectable()
export class StructuresSearchService {
  private index = 'structures';
  private readonly logger = new Logger(StructuresSearchService.name);

  constructor(private readonly elasticsearchService: ElasticsearchService) {}

  public async indexStructure(structure: StructureDocument): Promise<StructureDocument> {
    this.logger.debug(`indexStructure`);
    this.elasticsearchService.index<StructureSearchResult, StructureSearchBody>({
      index: this.index,
      id: structure._id,
      body: {
        structureName: structure.structureName,
        structureType: structure.structureType.value,
        structureId: structure._id,
        address: structure.address,
        description: structure.description,
      },
    });
    return structure;
  }

  public async createStructureIndex(): Promise<any> {
    this.logger.debug(`createStructureIndex`);
    // use custom analyzer with minimal_french stemmer to avoid default light_french stemmer problems (oullins -> oulin, "oull" not found)
    // don't use stopwords (le, la, de, son, etc.) for structureName (to find "maison de la", "le son du clic", etc.)
    return this.elasticsearchService.indices.create({
      index: this.index,
      body: {
        mappings: {
          dynamic_templates: [
            {
              strings: {
                match_mapping_type: 'string',
                mapping: {
                  type: 'text',
                  analyzer: 'homemade_french',
                  search_analyzer: 'homemade_french',
                },
              },
            },
          ],
          properties: {
            structureName: {
              type: 'text',
              analyzer: 'homemade_french_stopless',
              search_analyzer: 'homemade_french_stopless',
            },
          },
        },
        settings: es_settings_homemade_french,
      },
    });
  }

  public async dropIndex(): Promise<any> {
    this.logger.debug(`dropIndex`);
    if (
      (
        await this.elasticsearchService.indices.exists({
          index: this.index,
        })
      ).body
    ) {
      return this.elasticsearchService.indices.delete({
        index: this.index,
      });
    }
  }

  public async deleteIndexStructure(structure: StructureDocument): Promise<StructureDocument> {
    this.logger.debug(`deleteIndexStructure`);
    this.elasticsearchService.delete<StructureSearchResult, StructureSearchBody>({
      index: this.index,
      id: structure._id,
    });
    return structure;
  }

  public async refreshIndexStructure(): Promise<any> {
    this.logger.debug(`refreshIndexStructure`);
    return this.elasticsearchService.indices.refresh({
      index: this.index,
    });
  }

  public async search(searchString: string, fields?: string[]): Promise<StructureSearchBody[]> {
    this.logger.debug(`search ${searchString} | fields : ${fields}`);
    searchString = searchString ? searchString + '*' : '*';
    const { body } = await this.elasticsearchService.search<StructureSearchResult>({
      index: this.index,
      body: {
        from: 0,
        size: 200,
        query: {
          query_string: {
            analyze_wildcard: 'true',
            query: searchString,
            fields: fields ? fields : ['structureName^5', 'structureType^5', 'address.commune^10', 'description'],
            fuzziness: 'AUTO',
          },
        },
      },
    });
    const sortedHits = body.hits.hits.filter(function (elem) {
      return elem._score >= body.hits.max_score / 1.5;
    });
    return sortedHits.map((item) => item._source);
  }

  public async update(structure: Structure, id: string): Promise<any> {
    this.logger.debug('update');
    return this.elasticsearchService.update({
      index: this.index,
      id: id,
      body: {
        doc: {
          structureName: structure.structureName,
          structureType: structure.structureType,
          structureId: id,
          address: structure.address,
          description: structure.description,
        },
      },
    });
  }
}
