import { HttpService } from '@nestjs/axios';
import { HttpException, HttpStatus, Injectable, Logger } from '@nestjs/common';
import { InjectModel } from '@nestjs/mongoose';
import { Cron, CronExpression } from '@nestjs/schedule';
import axios, { AxiosResponse } from 'axios';
import * as ejs from 'ejs';
import * as _ from 'lodash';
import { DateTime } from 'luxon';
import { DocumentDefinition, FilterQuery, Model, Types } from 'mongoose';
import { Observable } from 'rxjs';
import { UnclaimedStructureDto } from '../../admin/dto/unclaimed-structure-dto';
import { CategoriesModule } from '../../categories/categories.module';
import { CategoriesAccompagnement } from '../../categories/schemas/categoriesAccompagnement.schema';
import { CategoriesFormations } from '../../categories/schemas/categoriesFormations.schema';
import { CategoriesFormationsModule } from '../../categories/schemas/categoriesFormationsModule.schema';
import { CategoriesOthers } from '../../categories/schemas/categoriesOthers.schema';
import { CategoriesFormationsService } from '../../categories/services/categories-formations.service';
import { MailerService } from '../../mailer/mailer.service';
import { PersonalOfferDocument } from '../../personal-offers/schemas/personal-offer.schema';
import { IUser } from '../../users/interfaces/user.interface';
import { User } from '../../users/schemas/user.schema';
import { UsersService } from '../../users/services/users.service';
import { PhotonResponse } from '../common/photonResponse.type';
import { depRegex } from '../common/regex';
import { StructureDto } from '../dto/structure.dto';
import { UpdateStructureDto } from '../dto/update-structure.dto';
import { CNFSStructure } from '../interfaces/cnfs-structure.interface';
import { Structure, StructureDocument } from '../schemas/structure.schema';
import { StructuresSearchService } from './structures-search.service';

@Injectable()
export class StructuresService {
  private readonly logger = new Logger(StructuresService.name);
  constructor(
    private readonly httpService: HttpService,
    private readonly userService: UsersService,
    private readonly mailerService: MailerService,
    private structuresSearchService: StructuresSearchService,
    private categoriesFormationsService: CategoriesFormationsService,
    @InjectModel(Structure.name) private structureModel: Model<StructureDocument>
  ) {}

  async initiateStructureIndex(): Promise<StructureDocument[]> {
    this.logger.log('Reset structures indexes');
    await this.structuresSearchService.dropIndex();
    await this.structuresSearchService.createStructureIndex();
    return this.populateES();
  }

  public fillFilters(filters: Array<any>): Promise<any[]>[] {
    return filters?.map(async (elem) => {
      const key = Object.keys(elem)[0];
      const modules = (await this.categoriesFormationsService.findOneComplete(elem[key])).modules;
      return modules.map((module) => {
        return { [elem[key]]: module.id };
      });
    });
  }

  async searchForStructures(text: string, filters?: Array<any>, fields?: string[]): Promise<StructureDocument[]> {
    const results = await this.structuresSearchService.search(text, fields);
    const ids = results.map((result) => result.structureId);
    let structures;
    let multipleFilters = filters ? filters.filter((elem) => Object.keys(elem)[0].length == 0) : null;
    filters = filters?.filter((elem) => Object.keys(elem)[0].length != 0);
    if (multipleFilters) {
      const filtersArrays = await Promise.all(this.fillFilters(multipleFilters));
      multipleFilters = [...filtersArrays];
    }
    if (!ids.length) {
      return [];
    }
    //we match ids from Elasticsearch with ids from mongoDB (and filters) and sort the result according to ElasticSearch order.
    if (filters?.length > 0 && multipleFilters?.length == 0) {
      structures = await this.structureModel
        .find({
          _id: { $in: ids },
          $and: [...this.parseFilter(filters), { deletedAt: { $exists: false }, accountVerified: true }],
        })
        .populate('personalOffers')
        .exec();
    } else if (filters?.length > 0 && multipleFilters?.length > 0) {
      structures = await this.structureModel
        .find({
          _id: { $in: ids },
          $or: [...this.parseFilter(multipleFilters)],
          $and: [...this.parseFilter(filters), { deletedAt: { $exists: false }, accountVerified: true }],
        })
        .populate('personalOffers')
        .exec();
    } else if (filters?.length == 0 && multipleFilters?.length > 0) {
      structures = await this.structureModel
        .find({
          _id: { $in: ids },
          $or: [...this.parseFilter(multipleFilters), { deletedAt: { $exists: false }, accountVerified: true }],
        })
        .populate('personalOffers')
        .exec();
    } else {
      structures = await this.structureModel
        .find({
          _id: { $in: ids },
          $and: [{ deletedAt: { $exists: false }, accountVerified: true }],
        })
        .populate('personalOffers')
        .exec();
    }

    return structures.sort((a, b) => ids.indexOf(a.id) - ids.indexOf(b.id));
  }

  public async create(email: string, structure: StructureDto): Promise<Structure> {
    const user = await this.userService.findOne(email);
    if (!user) {
      throw new HttpException('Invalid profile', HttpStatus.NOT_FOUND);
    }
    // Handle equipments
    // Value is forced to true because there is no verification needed now. Cannot be removed because of old data prod
    structure.accountVerified = true;
    let createdStructure = new this.structureModel(structure);
    createdStructure._id = Types.ObjectId();
    createdStructure = this.addEquipmentsCategories(createdStructure);
    await createdStructure.save();
    await this.getStructurePosition(createdStructure).then(async (position: StructureDocument) => {
      return this.structuresSearchService.indexStructure(
        await this.structureModel
          .findByIdAndUpdate(Types.ObjectId(createdStructure._id), {
            address: position.address,
            coord: position.coord,
          })
          .populate('personalOffers')
          .exec()
      );
    });
    user.structuresLink.push(createdStructure._id);
    user.save();

    // Send admin notification mail
    this.sendAdminStructureNotification(
      createdStructure,
      this.mailerService.config.templates.adminStructureCreate.ejs,
      this.mailerService.config.templates.adminStructureCreate.json
    );

    // Async treatment that can be done in backgorund, DO NOT await because it takes around 3sec to finish
    this.bindOneCNFSStructure(createdStructure.id);

    return createdStructure;
  }

  public async search(searchString: string, filters?: Array<any>): Promise<Structure[]> {
    if (searchString && filters) {
      return this.structureModel
        .find({
          $and: [
            ...this.parseFilter(filters),
            { $text: { $search: searchString }, deletedAt: { $exists: false }, accountVerified: true },
          ],
        })
        .populate('personalOffers')
        .exec();
    } else if (filters) {
      return this.structureModel
        .find({ $and: [{ $and: this.parseFilter(filters), deletedAt: { $exists: false }, accountVerified: true }] })
        .populate('personalOffers')
        .exec();
    } else {
      return this.structureModel
        .find({
          $and: [{ $or: [{ $text: { $search: searchString }, deletedAt: { $exists: false }, accountVerified: true }] }],
        })
        .populate('personalOffers')
        .exec();
    }
  }

  /**
   * Parse equipments. Ex : if structure doesn't have computers but nbComputer changes, then the key is added to equipmentsAndServices
   * @param structure
   * @returns Structure
   */
  public addEquipmentsCategories(structure: StructureDocument): StructureDocument {
    Logger.debug(`addEquipmentsCategories of ${structure.structureName}`, StructuresService.name);
    if (structure.nbComputers > 0) {
      structure.equipmentsAndServices.push('ordinateurs');
    }
    if (structure.nbPrinters > 0) {
      structure.equipmentsAndServices.push('imprimantes');
    }
    if (structure.nbNumericTerminal > 0) {
      structure.equipmentsAndServices.push('bornesNumeriques');
    }
    if (structure.nbScanners > 0) {
      structure.equipmentsAndServices.push('scanners');
    }
    if (structure.nbTablets > 0) {
      structure.equipmentsAndServices.push('tablettes');
    }
    // Remove duplicate if exists
    structure.equipmentsAndServices = [...new Set(structure.equipmentsAndServices)];
    return structure;
  }

  /**
   * Parse filter value from string to boolean
   * @param filters
   */
  private parseFilter(filters: Array<any>): Array<any> {
    return filters.map((filter) => {
      const key = Object.keys(filter)[0];
      if (filter[key] === 'True') {
        return { [key]: true };
      } else {
        return filter;
      }
    });
  }

  public async findAll(): Promise<StructureDocument[]> {
    this.logger.debug('findAll');
    const structures = await this.structureModel.find({ deletedAt: { $exists: false } }).exec();
    // Update structures coord and address before sending them
    await Promise.all(
      structures.map((structure: StructureDocument) => {
        // If structre has no address, add it
        if (!structure.address || structure.coord.length <= 0) {
          return this.getStructurePosition(structure).then((position: StructureDocument) => {
            this.structureModel
              .findByIdAndUpdate(Types.ObjectId(structure._id), { address: position.address, coord: position.coord })
              .exec();
          });
        }
      })
    );
    return this.structureModel.find({ deletedAt: { $exists: false }, accountVerified: true }).exec();
  }

  public async findAllUnclaimed(): Promise<UnclaimedStructureDto[]> {
    const structures = await this.structureModel
      .find({ deletedAt: { $exists: false } })
      .populate('personalOffers')
      .exec();
    const unclaimedStructures = [];
    await Promise.all(
      structures.map(async (structure: StructureDocument) => {
        if (!(await this.userService.isStructureClaimed(structure.id))) {
          unclaimedStructures.push({
            structureId: structure.id,
            structureName: structure.structureName,
            createdAt: structure.createdAt,
            updatedAt: structure.updatedAt,
          });
        }
      })
    );
    return unclaimedStructures;
  }

  public async populateES(): Promise<StructureDocument[]> {
    const structures = await this.structureModel
      .find({ deletedAt: { $exists: false } })
      .populate('personalOffers')
      .exec();
    await Promise.all(
      structures.map((structure: StructureDocument) => {
        this.structuresSearchService.indexStructure(structure);
      })
    );
    return structures;
  }

  public async findAllFormated(
    formationCategories: CategoriesFormations[],
    accompagnementCategories: CategoriesAccompagnement[],
    otherCategories: CategoriesOthers[]
  ): Promise<StructureDocument[]> {
    const structures = await this.structureModel
      .find({ deletedAt: { $exists: false } })
      .populate('personalOffers')
      .exec();
    // Update structures coord and address before sending them
    this.logger.debug('Find all formated structures, only returning structures who consented to data sharing');
    await Promise.all(
      structures.map((structure: StructureDocument) => {
        // If structure has no address, add it
        if (!structure.address || structure.coord.length <= 0) {
          return this.getStructurePosition(structure).then((position: StructureDocument) => {
            this.structureModel
              .findByIdAndUpdate(Types.ObjectId(structure._id), {
                address: position.address,
                coord: position.coord,
              })
              .populate('personalOffers')
              .exec();
          });
        }
      })
    );
    return (
      await this.structureModel
        .find({ deletedAt: { $exists: false }, accountVerified: true, $where: 'this.dataShareConsentDate != null' })
        .select('-_id -accountVerified -otherDescription -dataShareConsentDate')
        .exec()
    ).map((structure) => {
      // If structure has temp email, hide it
      if (this.hasTempMail(structure)) {
        structure.contactMail = null;
      }
      // Add referentiel
      structure.proceduresAccompaniment = this.mapModules(
        structure.proceduresAccompaniment,
        accompagnementCategories.find((category) => category.id === 'proceduresAccompaniment').modules
      );
      structure.labelsQualifications = this.mapModules(
        structure.labelsQualifications,
        otherCategories.find((category) => category.id === 'labelsQualifications').modules
      );
      structure.publics = this.mapModules(
        structure.publics,
        otherCategories.find((category) => category.id === 'publics').modules
      );
      structure.accessModality = this.mapModules(
        structure.accessModality,
        otherCategories.find((category) => category.id === 'accessModality').modules
      );
      structure.publicsAccompaniment = this.mapModules(
        structure.publicsAccompaniment,
        otherCategories.find((category) => category.id === 'publicsAccompaniment').modules
      );
      structure.equipmentsAndServices = this.mapModules(
        structure.equipmentsAndServices,
        otherCategories.find((category) => category.id === 'equipmentsAndServices').modules
      );
      structure.baseSkills = this.mapFormationModules(
        structure.baseSkills,
        formationCategories.find((category) => category.id === 'baseSkills').modules
      );
      structure.accessRight = this.mapFormationModules(
        structure.accessRight,
        formationCategories.find((category) => category.id === 'accessRight').modules
      );
      structure.socialAndProfessional = this.mapFormationModules(
        structure.socialAndProfessional,
        formationCategories.find((category) => category.id === 'socialAndProfessional').modules
      );
      structure.parentingHelp = this.mapFormationModules(
        structure.parentingHelp,
        formationCategories.find((category) => category.id === 'parentingHelp').modules
      );
      structure.digitalCultureSecurity = this.mapFormationModules(
        structure.digitalCultureSecurity,
        formationCategories.find((category) => category.id === 'digitalCultureSecurity').modules
      );
      return structure;
    });
  }
  public mapFormationModules(structureModule: string[], baseModule: CategoriesFormationsModule[]): string[] {
    if (structureModule.length === 0) {
      return [];
    }
    return structureModule.map((id) => {
      const formatedText = baseModule.find((module) => module.display_id === id)?.text;
      return formatedText ? formatedText : id;
    });
  }
  public mapModules(structureModule: string[], baseModule: CategoriesModule[]): string[] {
    if (structureModule.length === 0) {
      return [];
    }
    return structureModule.map((id) => {
      const formatedText = baseModule.find((module) => module.id === id)?.text;
      return formatedText ? formatedText : id;
    });
  }

  public async update(idStructure: string, updatedFields: UpdateStructureDto): Promise<Structure> {
    this.logger.debug(`Updating structure with id: ${idStructure}`);
    const oldStructure = await this.findOne(idStructure);
    const deepClone = Object.assign(oldStructure, updatedFields);

    if (!_.isEqual(oldStructure.address, updatedFields.address)) {
      await this.getStructurePosition(deepClone).then();
    }
    const parsedStructure = this.addEquipmentsCategories(deepClone);

    const result = await this.structureModel.findByIdAndUpdate(Types.ObjectId(idStructure), parsedStructure).exec();
    if (!result) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    const responseStructure = await this.findOne(idStructure);
    await this.structuresSearchService.update(responseStructure, idStructure);
    this.sendAdminStructureNotification(
      result,
      this.mailerService.config.templates.structureModificationNotification.ejs,
      this.mailerService.config.templates.structureModificationNotification.json
    );
    this.userService.removeOutdatedStructureFromArray(idStructure);

    return responseStructure;
  }

  public async findOne(idParam: string): Promise<StructureDocument> {
    return this.structureModel
      .findById(Types.ObjectId(idParam))
      .populate('personalOffers')
      .populate('structureType')
      .exec();
  }
  /**
   * Get structures positions and add marker corresponding to those positons on the map
   */
  private getStructurePosition(structure: Structure): Promise<Structure> {
    return new Promise((resolve, reject) => {
      this.getCoord(structure.address.numero, structure.address.street, structure.address.commune, 'photon').subscribe(
        (res) => {
          this.getCoord(
            structure.address.numero,
            structure.address.street,
            structure.address.commune,
            'photon-bal'
          ).subscribe((resbal) => {
            // check if photon-bal is more precise than photon ban
            if (resbal.data.features.length > 0) {
              resbal.data.features = resbal.data.features.sort((_a, b) => {
                return b.properties.housenumber ? 1 : -1;
              });
              const address = resbal.data.features[0];
              if (address && address.geometry) {
                structure.coord = address.geometry.coordinates;
              } else {
                this.logger.error(
                  `No coord found for: ${structure.address.numero} ${structure.address.street} ${structure.address.commune}`,
                  StructuresService.name
                );
                structure.coord = [];
              }
              resolve(structure);
              // else pick structure from photonban if it exists or throw error
            } else {
              if (res.data.features.length > 0) {
                const address = res.data.features[0];
                if (address && address.geometry) {
                  structure.coord = address.geometry.coordinates;
                } else {
                  this.logger.error(
                    `No coord found for: ${structure.address.numero} ${structure.address.street} ${structure.address.commune}`,
                    StructuresService.name
                  );
                  structure.coord = [];
                }
                resolve(structure);
              } else {
                this.logger.error(
                  `No structure found for: ${structure.address.numero} ${structure.address.street} ${structure.address.commune}`,
                  StructuresService.name
                );
              }
            }
          });
        },
        (err) => {
          this.logger.error(`Request error: ${err.config.url}`, 'StructureService');
          this.logger.error(err);
          reject(err);
        }
      );
    });
  }

  public async isClaimed(structureId: string, user: User): Promise<boolean> {
    const isStructureClaimed = await this.userService.isStructureClaimed(structureId);
    const isUserAlreadyClaimed = await this.userService.isUserAlreadyClaimedStructure(structureId, user.email);
    if (isStructureClaimed || isUserAlreadyClaimed) {
      return true;
    }
    return false;
  }

  /**
   * Search structure address based on data search WS
   * @param {searchQuery} data - Query address
   */
  public async searchAddress(data: {
    searchQuery: string;
  }): Promise<{ features: { geometry: any; type: string; properties: any }[] }> {
    const reqBan = `https://download.data.grandlyon.com/geocoding/photon/api?q=${data.searchQuery}&lang=fr&limit=500&osm_tag=:!construction&osm_tag=:!bus_stop`;

    const reqBal = `https://download.data.grandlyon.com/geocoding/photon-bal/api?q=${data.searchQuery}&lang=fr&limit=500&osm_tag=:!construction&osm_tag=:!bus_stop`;

    const requestGroup = (url): Promise<{ features: { geometry: any; type: string; properties: any }[] }> =>
      new Promise((resolve) => {
        this.logger.debug(`Search request: ${encodeURI(url)}`, 'StructureService');
        return this.httpService
          .request({
            url: encodeURI(url),
            method: 'GET',
            headers: { 'Content-Type': 'application/json' },
          })
          .subscribe(
            (reply) => {
              this.logger.debug(`Search request response length : ${reply.data.features.length}`, 'StructureService');
              reply.data.features = reply.data.features
                .filter((doc) => doc.properties.postcode && doc.properties.postcode.match(depRegex))
                .sort((_a, b) => {
                  return b.properties.housenumber ? 1 : -1;
                });
              return resolve(reply.data);
            },
            (err) => {
              this.logger.error(`Search - Request error: ${err.config.url}`, 'StructureService');
              this.logger.error(err);
            }
          );
      });

    const [reqBanRes, reqBalRes] = await Promise.all([requestGroup(reqBan), requestGroup(reqBal)]);
    const mergedArray = [...reqBalRes['features'], ...reqBanRes['features']];
    const duplicateFreeArray = _.unionWith(mergedArray, function (a, b) {
      return (
        // excludes structures from photon-BAN if they share postcode && street or name && housenumber
        //  checking for name and street existance asserts that we will not compare 2 undefined fields
        a.properties.postcode === b.properties.postcode &&
        ((a.properties.name &&
          (a.properties.name === b.properties.name || a.properties.name === b.properties.street)) ||
          (a.properties.street &&
            (a.properties.street === b.properties.name || a.properties.street === b.properties.street))) &&
        (a.properties.housenumber === b.properties.housenumber ||
          (!a.properties.housenumber && !b.properties.housenumber))
      );
    });

    return {
      features: duplicateFreeArray,
    };
  }

  /**
   * Count every value occurence of a given key
   * @param key structure key
   * @return [{id: 'key', count: 'value'}]
   */
  public async countByStructureKey(key: string, selected: { id: string; text: string }[]): Promise<any> {
    const uniqueElements = await this.structureModel.distinct(key).exec();
    return Promise.all(
      uniqueElements.map(async (value) => {
        const keyList: FilterQuery<DocumentDefinition<StructureDocument>>[] = [];
        keyList.push({
          [key]: { $elemMatch: { $eq: value } },
          deletedAt: { $exists: false },
          accountVerified: true,
        });
        if (selected && selected.length > 0) {
          for (const val of selected) {
            keyList.push({
              [val.text]: { $elemMatch: { $eq: val.id } },
              deletedAt: { $exists: false },
              accountVerified: true,
            });
          }
        }
        return {
          id: value,
          count: await this.structureModel
            .countDocuments({
              $and: keyList,
            })
            .exec(),
        };
      })
    );
  }

  public getCoord(
    numero: string,
    address: string,
    zipcode: string,
    scope: string
  ): Observable<AxiosResponse<PhotonResponse>> {
    const req =
      `https://download.data.grandlyon.com/geocoding/${scope}/api?q=` + numero + ' ' + address + ' ' + zipcode;
    this.logger.debug('Print getCoord' + req);

    return this.httpService.get(encodeURI(req));
  }

  /**
   * Set the structure to be deleted in 5 weeks
   * @param user IUser
   * @param structure StructureDocument
   */
  public async setToBeDeleted(user: IUser, structure: StructureDocument): Promise<Structure> {
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    structure.toBeDeletedAt = DateTime.local().plus({ weeks: 5 }).setZone('Europe/Paris').toString();
    structure.save();

    this.sendToBeDeletedNotification(user, structure);
    return structure;
  }

  /**
   * Send an email to structure owners (except the one who asked for the deletion) to inform the structure is will be deleted in 5 weeks
   * @param user User
   */
  public async sendToBeDeletedNotification(user: IUser, structure: StructureDocument): Promise<void> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureToBeDeleted.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureToBeDeleted.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      structureName: structure.structureName,
      name: user.name,
      surname: user.surname,
      toBeDeletedAt: DateTime.fromISO(structure.toBeDeletedAt.toISOString())
        .plus({ days: 1 })
        .toLocaleString(DateTime.DATE_SHORT),
    });
    const owners = await this.getOwners(structure._id);
    owners.forEach((owner) => {
      if (!owner._id.equals(user._id)) {
        this.mailerService.send(owner.email, jsonConfig.subject, html);
      }
    });
  }

  /**
   * Cancel the structure deletion when one of the structure owners had asked the structure for deletion
   * @param user IUser
   * @param structure StructureDocument
   */
  public async cancelDelete(user: IUser, structure: StructureDocument): Promise<Structure> {
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    if (!structure.deletedAt) {
      structure.toBeDeletedAt = null;
      structure.save();

      this.sendCancelDeleteNotification(user, structure);
    }
    return structure;
  }

  /**
   * Send an email to other structure owners to inform an owner cancelled the structure deletion
   * @param user User
   */
  public async sendCancelDeleteNotification(user: IUser, structure: StructureDocument): Promise<void> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureCancelDelete.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureCancelDelete.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      structureName: structure.structureName,
      name: user.name,
      surname: user.surname,
    });
    const owners = await this.getOwners(structure._id);
    owners.forEach((owner) => {
      if (!owner._id.equals(user._id)) {
        this.mailerService.send(owner.email, jsonConfig.subject, html);
      }
    });
  }

  public async deleteOne(structure: StructureDocument): Promise<Structure> {
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    this.structuresSearchService.deleteIndexStructure(structure);
    structure.structureType = null;
    if (structure.toBeDeletedAt) structure.toBeDeletedAt = null;
    structure.deletedAt = DateTime.local().setZone('Europe/Paris').toString();
    this.anonymizeStructure(structure).save();
    // Remove structure from owners (and check if there is a newly unattached user)
    const owners = await this.getOwners(structure._id);
    owners.forEach((owner) => {
      this.userService.removeFromStructureLinked(owner, structure._id);
    });
    this.sendAdminStructureNotification(
      structure,
      this.mailerService.config.templates.structureDeletionNotification.ejs,
      this.mailerService.config.templates.structureDeletionNotification.json
    );
    return structure;
  }

  public async sendAdminStructureNotification(
    structure: StructureDocument,
    templateLocation: any,
    jsonConfigLocation: any,
    user: any = null
  ) {
    const uniqueAdminEmails = [...new Set((await this.userService.getAdmins()).map((admin) => admin.email))].map(
      (item) => {
        return { email: item };
      }
    );

    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(templateLocation);
    const jsonConfig = this.mailerService.loadJsonConfig(jsonConfigLocation);
    const html = await ejs.renderFile(ejsPath, {
      config,
      id: structure ? structure._id : 0,
      structureName: structure ? structure.structureName : '',
      structureAddress: structure
        ? `${structure.address.numero || ''} ${structure.address.street} ${structure.address.commune}`
        : '',
      structureDescription: structure ? structure.otherDescription : '',
      user: user,
    });
    this.mailerService.send(uniqueAdminEmails, jsonConfig.subject, html);
  }

  private anonymizeStructure(structure: StructureDocument): StructureDocument {
    structure.contactPhone = '';
    structure.contactMail = '';
    structure.facebook = null;
    structure.twitter = null;
    structure.instagram = null;
    structure.website = '';
    return structure;
  }

  @Cron(CronExpression.EVERY_DAY_AT_4AM)
  public async structuresTasksProcess(): Promise<void> {
    this.logger.debug('structuresTasksProcess');

    await this.processToBeDeletedStructures().catch((error) => {
      this.logger.error(error);
    });

    await this.processOutdatedStructures().catch((error) => {
      this.logger.error(error);
    });
  }

  private async processToBeDeletedStructures(): Promise<void> {
    this.logger.debug('processToBeDeletedStructures');

    const structures = await this.structureModel
      .find()
      .where('toBeDeletedAt')
      .lte(DateTime.local())
      .where('deletedAt')
      .exists(false)
      .exec();

    structures.forEach((structure) => {
      this.logger.debug(`delete structure : ${structure.structureName} (${structure._id})`);
      this.deleteOne(structure);
    });
  }

  private async processOutdatedStructures(): Promise<void> {
    this.logger.debug('processOutdatedStructures');
    const OUTDATED_MONTH_TO_CHECK = 6;
    const structureList = await this.findAll();
    // Get outdated structures
    const filteredList = structureList.filter((structure) => {
      const updatedDate = DateTime.fromISO(structure.updatedAt.toISOString());
      const diff = updatedDate.diffNow('month').months;
      if (Math.abs(diff) > OUTDATED_MONTH_TO_CHECK) {
        this.logger.debug(`found outdated structure : ${structure.structureName}`);
        return true;
      }
    });
    // Get owners of outdated structures
    const ownerList = await Promise.all(
      filteredList.map(async (structure) => {
        const owner = await this.userService.isStructureClaimed(structure._id.toString());
        if (owner) {
          return { structure: structure, owner: owner };
        }
      })
    );
    // Send email if possible and update user
    ownerList
      .filter((x) => x != undefined)
      .forEach(async (data) => {
        this.userService.findOne(data.owner.email).then((user) => {
          // If mail is already sent, do not resend
          if (user.structureOutdatedMailSent.includes(data.structure._id)) {
            return;
          } else {
            this.sendOutdatedEmailToUser(data.owner.email, data.structure.structureName, data.structure._id);
            user.structureOutdatedMailSent.push(data.structure._id);
            user.save();
          }
        });
      });
  }

  /**
   * Send an email to prevent outdated
   * a new account.
   * @param user User
   */
  private async sendOutdatedEmailToUser(userEmail: string, structureName: string, id: string): Promise<any> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureOutdatedInfo.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureOutdatedInfo.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      name: structureName,
      id: id,
    });
    this.mailerService.send(userEmail, jsonConfig.subject, html);
  }

  /**
   * Send an email to structure owners in order to accept or decline a join request
   * @param user User
   */
  public async sendStructureJoinRequest(user: IUser, structure: StructureDocument): Promise<void> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureJoinRequest.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureJoinRequest.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      structureName: structure.structureName,
      name: user.name,
      surname: user.surname,
      id: structure._id,
      userId: user._id,
    });
    const owners = await this.getOwners(structure._id);
    owners.forEach((owner) => {
      if (!owner._id.equals(user._id)) {
        this.mailerService.send(owner.email, jsonConfig.subject, html);
      }
    });
  }

  private async getOwners(structureId: string): Promise<IUser[]> {
    // Get owners of outdated structures
    return this.userService.getStructureOwners(structureId);
  }

  public async updateAccountVerified(idStructure: string): Promise<Structure> {
    const structureLinked = await this.findOne(idStructure);
    const structure = new this.structureModel(structureLinked);
    if (!structure) {
      throw new HttpException('Invalid structure', HttpStatus.NOT_FOUND);
    }
    structure.accountVerified = true;
    structure.save();
    return structure;
  }

  public async findWithOwners(
    idStructure: string,
    emailUser: string
  ): Promise<{ structure: Structure; owners: IUser[] }> {
    const structure = await this.findOne(idStructure);
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    const owners =
      typeof emailUser == 'undefined'
        ? await this.userService.getStructureOwnersDetails(structure._id, '')
        : await this.userService.getStructureOwnersDetails(structure._id, emailUser);
    return { structure: structure, owners: owners };
  }

  public async reportStructureError(structureId: string, content: string) {
    const structure = await this.findOne(structureId);
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    const owners = await this.userService.getStructureOwnersDetails(structure._id, '');
    const admins = await this.userService.getAdmins();
    const emails = owners.map((owner) => owner.email).concat(admins.map((admin) => admin.email));
    const uniqueEmails = [...new Set(emails)];

    const emailsObject: { email: string }[] = uniqueEmails.map((item) => {
      return { email: item };
    });

    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureErrorReport.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureErrorReport.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      content: content,
      id: structure._id,
      structureName: structure.structureName,
    });
    this.mailerService.send(emailsObject, jsonConfig.subject, html);
  }

  private hasTempMail(structure: Structure): boolean {
    return structure.contactMail === 'unknown@unknown.com';
  }

  public async getAllUserCompletedStructures(users: IUser[]): Promise<any[]> {
    return Promise.all(
      users.map(async (user) => {
        return {
          id: user._id,
          createdAt: user.createdAt ? user.createdAt.toLocaleDateString() : '',
          surname: user.surname,
          name: user.name,
          email: user.email,
          phone: user.phone,
          job: user.job,
          employer: user.employer,
          structures: await Promise.all(
            user.structuresLink.map(async (id) => {
              return this.findOne(id.toHexString());
            })
          ),
        };
      })
    );
  }

  /**
   * Retrieve all structure pending for data.grandlyon consent sharing.
   * Case 1 : null value stands for 'no'
   * Case 2: a date stands for 'yes'
   * Case 3: no key in document when there is no reply to consent
   * @param users {IUser}
   * @returns {Structure[]}
   */
  public async getAllDataConsentPendingStructures(users: IUser): Promise<Structure[]> {
    const data = await Promise.all(
      users.structuresLink.map(async (structureId: Types.ObjectId) => {
        const structure = await this.findOne(structureId.toHexString());
        if (structure && structure.dataShareConsentDate === undefined) {
          return structure;
        }
      })
    );
    return data.filter((value) => value);
  }

  /**
   * Add the personal offer to the structure
   * @param structureId string
   * @param personalOfferDocument PersonalOfferDocument
   * @returns {Structure} structurew with personal offer added
   */
  public async addPersonalOffer(
    structureId: string,
    personalOfferDocument: PersonalOfferDocument
  ): Promise<StructureDocument> {
    const structure: StructureDocument = await this.findOne(structureId);
    if (!structure) {
      throw new HttpException('Structure not found for the personal offer attachment', HttpStatus.BAD_REQUEST);
    }
    if (
      structure.personalOffers &&
      structure.personalOffers.findIndex((personalOffer) => personalOffer._id === personalOfferDocument._id) > -1
    ) {
      throw new HttpException('Personal offer already exist in the structure', HttpStatus.BAD_REQUEST);
    } else {
      structure.personalOffers.push(personalOfferDocument);
    }
    await structure.save();
    return structure;
  }

  /**
   * Fetch espace coop and filter structures in 69XXX and structures with courriel and phone
   * @returns CNFSStructure[]
   */
  public async getCNFSStructures(): Promise<Array<CNFSStructure>> {
    this.logger.debug(`get CNFS structures`);
    const response: AxiosResponse<Array<CNFSStructure>> = await axios.get(
      'https://api.conseiller-numerique.gouv.fr/permanences'
    );

    const filteredData = response.data.filter((structure) => {
      if (structure.code_postal.substring(0, 2) === '69' && (structure.courriel || structure.telephone))
        return structure;
    });
    this.logger.debug(`${filteredData.length} CNFS rhone structures containing found`);
    return filteredData;
  }

  /**
   * Compare Resin & CNFS structures with email and phone, returns the structure when mathing
   */
  public findMatchingStructure(resinStructure: StructureDocument, CNFSData: CNFSStructure[]): CNFSStructure | null {
    // This check is mandatory because if resin phone number is '' we should return FALSE instead of UNDEFINED
    const resinPhoneNumber = resinStructure.contactPhone?.split(' ').join('') || false;
    return CNFSData.find(
      (structure) =>
        structure.courriel === resinStructure.contactMail ||
        structure.telephone?.replace('+33', '0') === resinPhoneNumber
    );
  }

  /**
   * When a structure is created, fetch espace coop and tries to find a match to bind CNFSid to the resin structure.
   * @param resinStructureId string
   */
  public async bindOneCNFSStructure(resinStructureId: string): Promise<string> {
    this.logger.debug(`bind CNFS id to one resin structure`);
    const resinStructure = await this.findOne(resinStructureId);
    const CNFSData = await this.getCNFSStructures();
    if (!resinStructure) return 'Could not find structure';

    const mathingStructure = this.findMatchingStructure(resinStructure, CNFSData);
    if (mathingStructure) {
      // if a corresponing structure has been found but its email is different, we probably will need to update it here.
      await this.setCNFSid(resinStructure.id, mathingStructure.id);
      return 'A match has been found, updating CNFSid';
    }
    this.logger.debug(`No match found`);
    return 'No match found';
  }

  /**
   * Fetch espace coop. If any structure matches, assign its CNFSid in our DB, if not assign ''
   * @returns "${number} of structures affected"
   */
  @Cron(CronExpression.EVERY_WEEK)
  public async bindCNFSids(): Promise<string> {
    this.logger.debug(`bind CNFS ids to resin structures`);
    const CNFSData = await this.getCNFSStructures();
    const resinStructures = await this.findAll();
    this.logger.debug(`${resinStructures.length} resin structures found`);

    let matchCount = 0;
    let dismatchCount = 0;

    for (const resinStructure of resinStructures) {
      const mathingStructure = this.findMatchingStructure(resinStructure, CNFSData);
      if (mathingStructure) {
        await this.setCNFSid(resinStructure.id, mathingStructure.id);
        matchCount++;
        // if a corresponing structure has been found but its email is different, we probably will need to update it here.
      } else {
        await this.setCNFSid(resinStructure.id, '');
        dismatchCount++;
      }
    }
    this.logger.debug(`${matchCount} out of ${dismatchCount} structures affected with a idCNFS`);
    return `${matchCount} structures affected`;
  }

  /**
   * Update the idCNFS of a structure with its value of ""
   */
  public async setCNFSid(idStructure: string, idCNFS: string): Promise<Structure> {
    this.logger.debug(`A resin structure '${idStructure}' matched with a CNFS one, updating its idCNFS '${idCNFS}'`);
    const oldStructure = await this.findOne(idStructure);
    const deepClone = Object.assign(oldStructure, { idCNFS });
    return this.structureModel.findByIdAndUpdate(Types.ObjectId(idStructure), deepClone).exec();
  }
}
