import { HttpService } from '@nestjs/axios';
import {
  Body,
  Controller,
  Delete,
  Get,
  HttpException,
  HttpStatus,
  Logger,
  Param,
  Patch,
  Post,
  Put,
  Query,
  Request,
  UseGuards,
} from '@nestjs/common';
import { ApiOperation, ApiParam, ApiTags } from '@nestjs/swagger';
import { Types } from 'mongoose';
import { JwtAuthGuard } from '../auth/guards/jwt-auth.guard';
import { CategoriesAccompagnementService } from '../categories/services/categories-accompagnement.service';
import { CategoriesFormationsService } from '../categories/services/categories-formations.service';
import { CategoriesOthersService } from '../categories/services/categories-others.service';
import { CreateTempUserDto } from '../temp-user/dto/create-temp-user.dto';
import { TempUserService } from '../temp-user/temp-user.service';
import { Roles } from '../users/decorators/roles.decorator';
import { IsStructureOwnerGuard } from '../users/guards/isStructureOwner.guard';
import { RolesGuard } from '../users/guards/roles.guard';
import { User } from '../users/schemas/user.schema';
import { IUser } from '../users/interfaces/user.interface';
import { UsersService } from '../users/services/users.service';
import { depRegex } from './common/regex';
import { CreateStructureDto } from './dto/create-structure.dto';
import { QueryStructure } from './dto/query-structure.dto';
import { UpdateStructureDto } from './dto/update-structure.dto';
import { Structure, StructureDocument } from './schemas/structure.schema';
import { StructuresService } from './services/structures.service';

@ApiTags('structures')
@Controller('structures')
export class StructuresController {
  private readonly logger = new Logger(StructuresController.name);
  constructor(
    private readonly httpService: HttpService,
    private readonly structureService: StructuresService,
    private readonly userService: UsersService,
    private readonly tempUserService: TempUserService,
    private readonly categoriesFormationsService: CategoriesFormationsService,
    private readonly categoriesOthersService: CategoriesOthersService,
    private readonly categoriesAccompagnementService: CategoriesAccompagnementService
  ) {}

  /**
   * Return points of given town exist.
   * @param zipcode
   * @returns Array of points
   */
  @Get('coordinates/:zipcode')
  @ApiParam({ name: 'zipcode', type: String, required: true })
  public async getCoordinates(@Param('zipcode') city: string): Promise<any> {
    return this.httpService
      .get(encodeURI(`https://download.data.grandlyon.com/geocoding/photon/api?q=${city}`))
      .toPromise()
      .then(async (res) => res.data.features)
      .then((data) =>
        data.filter(
          (cityPoint) =>
            cityPoint.properties.city?.toLowerCase().includes(city.toLowerCase()) &&
            cityPoint.properties.postcode.match(depRegex)
        )
      )
      .then((data) => data.map((filteredCityPoint) => filteredCityPoint.geometry.coordinates));
  }

  @Post()
  public async create(@Body() createStructureDto: CreateStructureDto): Promise<Structure> {
    Logger.debug(`create | ${createStructureDto.structure.structureName}`, StructuresController.name);
    return this.structureService.create(createStructureDto.userEmail, createStructureDto.structure);
  }

  @Post('search')
  public async search(@Query() query: QueryStructure, @Body() body): Promise<Structure[]> {
    return this.structureService.searchForStructures(query.query, body ? body.filters : null);
  }

  @Post('searchByName')
  public async searchByName(@Query() query: QueryStructure, @Body() body): Promise<Structure[]> {
    return this.structureService.searchForStructures(query.query, body ? body.filters : null, ['structureName']);
  }

  @Post('resetSearchIndex')
  @UseGuards(JwtAuthGuard, RolesGuard)
  @Roles('admin')
  public async resetES(): Promise<StructureDocument[]> {
    return this.structureService.initiateStructureIndex();
  }

  @Put('updateAfterOwnerVerify/:id')
  public async updateAfterOwnerVerify(@Param('id') id: string): Promise<Structure> {
    Logger.debug(`updateAfterOwnerVerify | structure ${id}`, StructuresController.name);
    return this.structureService.updateAccountVerified(id);
  }

  @ApiOperation({ summary: 'Patch one value or more of a structure' })
  @Patch(':id')
  @UseGuards(JwtAuthGuard, IsStructureOwnerGuard)
  @Roles('admin')
  public async update(@Param('id') id: string, @Body() body: UpdateStructureDto): Promise<Structure> {
    return this.structureService.update(id, body);
  }

  @Get()
  public async findAll(): Promise<Structure[]> {
    this.logger.debug('findAll');
    return this.structureService.findAll();
  }

  @Get('formated')
  public async findAllFormated(): Promise<Structure[]> {
    const formationCategories = await this.categoriesFormationsService.findAll();
    const accompagnementCategories = await this.categoriesAccompagnementService.findAll();
    const otherCategories = await this.categoriesOthersService.findAll();
    return this.structureService.findAllFormated(formationCategories, accompagnementCategories, otherCategories);
  }

  @Post(':id/isClaimed')
  public async isClaimed(@Param('id') id: string, @Body() user?: User): Promise<boolean> {
    return this.structureService.isClaimed(id, user);
  }

  @Post(':id/claim')
  public async claim(@Param('id') idStructure: string, @Body() user: User): Promise<Types.ObjectId[]> {
    const structure = await this.structureService.findOne(idStructure);
    return this.userService.updateStructureLinkedClaim(user.email, idStructure, structure);
  }

  @Post('address')
  public async searchAddress(@Body() data: { searchQuery: string }) {
    return this.structureService.searchAddress(data);
  }

  @Get(':id')
  public async find(@Param('id') id: string) {
    const result = await this.structureService.findOne(id);
    if (!result || result.deletedAt) {
      throw new HttpException('Structure does not exist', HttpStatus.NOT_FOUND);
    } else {
      return result;
    }
  }

  @Post(':id/withOwners')
  public async findWithOwners(@Param('id') id: string, @Body() data: { emailUser: string }) {
    return this.structureService.findWithOwners(id, data.emailUser);
  }

  @Get(':id/tempUsers')
  public async getTempUsers(@Param('id') id: string) {
    return this.tempUserService.getStructureTempUsers(id);
  }

  @Delete(':id')
  @UseGuards(JwtAuthGuard, IsStructureOwnerGuard)
  @ApiParam({ name: 'id', type: String, required: true })
  public async delete(@Request() req, @Param('id') id: string) {
    const structure = await this.structureService.findOne(id);
    if (!structure) {
      throw new HttpException('Invalid Structure', HttpStatus.NOT_FOUND);
    }
    const otherOwners: IUser[] = (await this.userService.getStructureOwners(id)).filter((owner) => {
      return !owner._id.equals(req.user._id);
    });
    if (otherOwners.length) {
      return this.structureService.setToBeDeleted(req.user, structure);
    } else {
      return this.structureService.deleteOne(structure);
    }
  }

  @Post(':id/cancelDelete')
  @UseGuards(JwtAuthGuard, IsStructureOwnerGuard)
  @ApiParam({ name: 'id', type: String, required: true })
  public async cancelDelete(@Request() req, @Param('id') id: string): Promise<Structure> {
    const structure = await this.structureService.findOne(id);
    if (!structure) {
      throw new HttpException('Invalid Structure', HttpStatus.NOT_FOUND);
    }
    return this.structureService.cancelDelete(req.user, structure);
  }

  @Post(':id/addOwner')
  @UseGuards(JwtAuthGuard, IsStructureOwnerGuard)
  @Roles('admin')
  @ApiParam({ name: 'id', type: String, required: true })
  public async addOwner(@Param('id') id: string, @Body() user: CreateTempUserDto): Promise<any> {
    // Get structure name
    const structure = await this.structureService.findOne(id);
    if (!structure) {
      throw new HttpException('Invalid Structure', HttpStatus.NOT_FOUND);
    }
    user.pendingStructuresLink = [Types.ObjectId(id)];
    // If user already exist, use created account
    if (await this.userService.verifyUserExist(user.email)) {
      return this.userService.updateStructureLinked(user.email, id);
    }
    // If temp user exist, update it
    if (await this.tempUserService.findOne(user.email)) {
      return this.tempUserService.updateStructureLinked(user);
    }
    // If not, create
    return this.tempUserService.create(user, structure.structureName);
  }

  @Post(':id/join')
  @ApiParam({ name: 'id', type: String, required: true })
  public async join(@Param('id') id: string, @Body() user: User): Promise<void> {
    // Get structure name
    const structure = await this.structureService.findOne(id);
    if (!structure) {
      throw new HttpException('Invalid Structure', HttpStatus.NOT_FOUND);
    }
    // Get user and add structure to user
    const userFromDb = await this.userService.findOne(user.email);
    if (!userFromDb) {
      throw new HttpException('Invalid User', HttpStatus.NOT_FOUND);
    }
    await this.userService.updateStructureLinked(userFromDb.email, id);
    // Send structure owners an email
    this.structureService.sendStructureJoinRequest(userFromDb, structure);
  }

  @Delete(':id/owner/:userId')
  @UseGuards(JwtAuthGuard, IsStructureOwnerGuard)
  @ApiParam({ name: 'id', type: String, required: true })
  @ApiParam({ name: 'userId', type: String, required: true })
  public async removeOwner(@Param('id') id: string, @Param('userId') userId: string): Promise<void> {
    // Get structure
    const structure = await this.structureService.findOne(id);
    if (!structure) {
      throw new HttpException('Invalid Structure', HttpStatus.NOT_FOUND);
    }
    // Get user
    const userFromDb = await this.userService.findById(userId);
    if (!userFromDb || !userFromDb.structuresLink.includes(Types.ObjectId(id))) {
      throw new HttpException('Invalid User', HttpStatus.NOT_FOUND);
    }
    this.userService.removeFromStructureLinked(userFromDb, id);
  }

  @Delete(':id/tempUser/:userId')
  @UseGuards(JwtAuthGuard, IsStructureOwnerGuard)
  @ApiParam({ name: 'id', type: String, required: true })
  @ApiParam({ name: 'userId', type: String, required: true })
  public async removeTempUser(@Param('id') id: string, @Param('userId') userId: string): Promise<void> {
    // Get structure
    const structure = await this.structureService.findOne(id);
    if (!structure) {
      throw new HttpException('Invalid Structure', HttpStatus.NOT_FOUND);
    }
    // Get temp user
    const userFromDb = await this.tempUserService.findById(userId);
    if (!userFromDb || !userFromDb.pendingStructuresLink.includes(Types.ObjectId(id))) {
      throw new HttpException('Invalid temp user', HttpStatus.NOT_FOUND);
    }
    this.tempUserService.removeFromStructureLinked(userFromDb.email, id);
  }

  @Post('reportStructureError')
  public async reportStructureError(@Body() data: { structureId: string; content: string }): Promise<void> {
    return this.structureService.reportStructureError(data.structureId, data.content);
  }
}
