import { HttpModule } from '@nestjs/axios';
import { HttpStatus } from '@nestjs/common';
import { getModelToken } from '@nestjs/mongoose';
import { Test, TestingModule } from '@nestjs/testing';
import { Types } from 'mongoose';
import { CategoriesFormationsServiceMock } from '../../../test/mock/services/categoriesFormations.mock.service';
import { CategoriesFormationsService } from '../../categories/services/categories-formations.service';
import { ConfigurationModule } from '../../configuration/configuration.module';
import { MailerService } from '../../mailer/mailer.service';
import { ParametersService } from '../../parameters/parameters.service';
import { Parameters } from '../../parameters/schemas/parameters.schema';
import { SearchModule } from '../../search/search.module';
import { Structure } from '../../structures/schemas/structure.schema';
import { StructuresSearchService } from '../../structures/services/structures-search.service';
import { StructuresService } from '../../structures/services/structures.service';
import { TempUser } from '../../temp-user/temp-user.schema';
import { TempUserService } from '../../temp-user/temp-user.service';
import { EmailChangeDto } from '../dto/change-email.dto';
import { PasswordChangeDto } from '../dto/change-password.dto';
import { DescriptionDto } from '../dto/description.dto';
import { ProfileDto } from '../dto/profile.dto';
import { UpdateDetailsDto } from '../dto/update-details.dto';
import { User } from '../schemas/user.schema';
import { EmployerService } from '../services/employer.service';
import { JobsService } from '../services/jobs.service';
import { UsersService } from '../services/users.service';
import { PasswordResetApplyDto } from './../dto/reset-password-apply.dto';
import { PasswordResetDto } from './../dto/reset-password.dto';
import { UsersController } from './users.controller';
import { usersMockData } from '../../../test/mock/data/users.mock.data';
import { CreateUserDto } from '../dto/create-user.dto';

describe('UsersController', () => {
  let controller: UsersController;

  const employerServiceMock = {
    findByName: jest.fn(),
  };

  const jobServiceMock = {
    findByName: jest.fn(),
  };

  const userServiceMock = {
    changeUserEmail: jest.fn(),
    changeUserPassword: jest.fn(),
    create: jest.fn(),
    deleteOne: jest.fn(),
    findById: jest.fn(),
    findOne: jest.fn(),
    isStructureClaimed: jest.fn(),
    sendResetPasswordEmail: jest.fn(),
    updateDescription: jest.fn(),
    updateStructureLinkedClaim: jest.fn(),
    updateUserDetails: jest.fn(),
    updateUserProfile: jest.fn(),
    validatePasswordResetToken: jest.fn(),
    validateUser: jest.fn(),
    verifyAndUpdateUserEmail: jest.fn(),
    verifyUserExist: jest.fn(),
  };

  const structureServiceMock = {
    deleteOne: jest.fn(),
    findOne: jest.fn(),
    getAllDataConsentPendingStructures: jest.fn(),
    sendAdminStructureNotification: jest.fn(),
  };

  const tempUserServiceMock = {
    delete: jest.fn(),
    findOne: jest.fn(),
  };

  beforeEach(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [ConfigurationModule, HttpModule, SearchModule],
      providers: [
        UsersService,
        StructuresSearchService,
        MailerService,
        ParametersService,
        {
          provide: StructuresService,
          useValue: structureServiceMock,
        },
        {
          provide: UsersService,
          useValue: userServiceMock,
        },
        {
          provide: TempUserService,
          useValue: tempUserServiceMock,
        },
        {
          provide: EmployerService,
          useValue: employerServiceMock,
        },
        {
          provide: JobsService,
          useValue: jobServiceMock,
        },
        {
          provide: CategoriesFormationsService,
          useValue: CategoriesFormationsServiceMock,
        },
        {
          provide: getModelToken('TempUser'),
          useValue: TempUser,
        },
        {
          provide: getModelToken('Structure'),
          useValue: Structure,
        },
        {
          provide: getModelToken('User'),
          useValue: User,
        },
        {
          provide: getModelToken('Parameters'),
          useValue: Parameters,
        },
      ],
      controllers: [UsersController],
    }).compile();
    afterEach(() => {
      jest.clearAllMocks();
    });

    controller = module.get<UsersController>(UsersController);
  });

  it('should be defined', () => {
    expect(controller).toBeDefined();
  });

  describe('getProfile', () => {
    it('should return user', () => {
      const user = { _id: '36', email: 'a@a.com' };
      const result = controller.getProfile({ user: user });
      expect(result).toEqual(user);
    });
  });

  describe('setProfile', () => {
    it('should return employer does not exist', async () => {
      const profile: ProfileDto = {
        employerName: 'Metro',
        jobName: 'Dev',
      };
      employerServiceMock.findByName.mockResolvedValueOnce(null);
      try {
        await controller.setProfile({ user: { _id: '36', email: 'a@a.com' } }, profile);
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toBe('Employer does not exist');
        expect(e.status).toBe(400);
      }
    });
    it('should return job does not exist', async () => {
      const profile: ProfileDto = {
        employerName: 'Metro',
        jobName: 'Dev',
      };
      employerServiceMock.findByName.mockResolvedValueOnce({
        _id: Types.ObjectId('6231aefe76598527c8d0b5a7'),
        name: 'CAF',
        validated: true,
      });
      jobServiceMock.findByName.mockResolvedValue(null);
      try {
        await controller.setProfile({ user: { _id: '36', email: 'a@a.com' } }, profile);
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toBe('Job does not exist');
        expect(e.status).toBe(400);
      }
    });
    it('should update profile', async () => {
      const profile: ProfileDto = {
        employerName: 'Metro',
        jobName: 'Dev',
      };
      employerServiceMock.findByName.mockResolvedValueOnce({
        _id: Types.ObjectId('6231aefe76598527c8d0b5a7'),
        name: 'CAF',
        validated: true,
      });
      jobServiceMock.findByName.mockResolvedValue({
        _id: Types.ObjectId('6231aefe76598527c8d0b5be'),
        name: 'Dev',
        validated: false,
        hasPersonalOffer: false,
      });
      userServiceMock.findOne.mockResolvedValueOnce({
        _id: Types.ObjectId('620e4ffa72389045b02ac854'),
        structuresLink: [Types.ObjectId('620e5236f25755550cb86dfd')],
        structureOutdatedMailSent: [],
        pendingStructuresLink: [Types.ObjectId('61e926192ac971550065e275')],
        phone: '06 06 06 06 06',
        newEmail: null,
        changeEmailToken: null,
        role: 1,
        resetPasswordToken: null,
        surname: 'ADMIN',
        name: 'Admin',
        validationToken: null,
        emailVerified: true,
        email: 'admin@admin.com',
        password: '$2a$12$vLQjJ9zAWyUwiXLeQDa6w.XzrlgPBhw.2GWrjog/yuEjIaZnQwmZu',
        __v: 11,
        employer: Types.ObjectId('6231aefe76598527c8d0b5a7'),
        job: Types.ObjectId('6231aefe76598527c8d0b5be'),
      });
      const reply = await controller.setProfile({ user: { _id: '36', email: 'a@a.com' } }, profile);
      expect(reply).toBeTruthy();
    });
  });

  describe('update details', () => {
    it('should update details', async () => {
      const newDetails: UpdateDetailsDto = {
        name: 'Jean',
        surname: 'Jacque',
        phone: '0605040302',
      };
      userServiceMock.updateUserDetails.mockResolvedValueOnce({
        _id: Types.ObjectId('620e4ffa72389045b02ac854'),
        structuresLink: [Types.ObjectId('620e5236f25755550cb86dfd')],
        structureOutdatedMailSent: [],
        pendingStructuresLink: [Types.ObjectId('61e926192ac971550065e275')],
        phone: '06 06 06 06 06',
        newEmail: null,
        changeEmailToken: null,
        role: 1,
        resetPasswordToken: null,
        surname: 'ADMIN',
        name: 'Admin',
        validationToken: null,
        emailVerified: true,
        email: 'admin@admin.com',
        password: '$2a$12$vLQjJ9zAWyUwiXLeQDa6w.XzrlgPBhw.2GWrjog/yuEjIaZnQwmZu',
        __v: 11,
        employer: Types.ObjectId('6231aefe76598527c8d0b5a7'),
        job: Types.ObjectId('6231aefe76598527c8d0b5be'),
      });
      const reply = await controller.updateDetails({ user: { _id: '36', email: 'a@a.com' } }, newDetails);
      expect(reply).toBeTruthy();
    });
  });

  describe('create', () => {
    it('should create user without structure', async () => {
      const userCreateSpyer = jest.spyOn(userServiceMock, 'create');
      const structureFindOneSpyer = jest.spyOn(structureServiceMock, 'findOne');
      const updateStructureLinkedClaimSpyer = jest.spyOn(userServiceMock, 'updateStructureLinkedClaim');
      const sendAdminStructureNotificationSpyer = jest.spyOn(structureServiceMock, 'sendAdminStructureNotification');
      const tempUserFindOneSpyer = jest.spyOn(tempUserServiceMock, 'findOne');
      const tempUserDeleteSpyer = jest.spyOn(tempUserServiceMock, 'delete');

      const createUserDto = new CreateUserDto();
      createUserDto.pendingStructuresLink = [];
      userServiceMock.create.mockResolvedValueOnce(usersMockData[0]);
      const result = await controller.create(createUserDto);

      expect(userCreateSpyer).toBeCalledTimes(1);
      expect(structureFindOneSpyer).toBeCalledTimes(0);
      expect(updateStructureLinkedClaimSpyer).toBeCalledTimes(0);
      expect(sendAdminStructureNotificationSpyer).toBeCalledTimes(0);
      expect(tempUserFindOneSpyer).toBeCalledTimes(1);
      expect(tempUserDeleteSpyer).toBeCalledTimes(0);
      expect(result).toEqual(usersMockData[0]);
    });

    it('should create user with structure', async () => {
      const userCreateSpyer = jest.spyOn(userServiceMock, 'create');
      const structureFindOneSpyer = jest.spyOn(structureServiceMock, 'findOne');
      const updateStructureLinkedClaimSpyer = jest.spyOn(userServiceMock, 'updateStructureLinkedClaim');
      const sendAdminStructureNotificationSpyer = jest.spyOn(structureServiceMock, 'sendAdminStructureNotification');
      const tempUserFindOneSpyer = jest.spyOn(tempUserServiceMock, 'findOne');
      const tempUserDeleteSpyer = jest.spyOn(tempUserServiceMock, 'delete');

      const createUserDto = new CreateUserDto();
      createUserDto.pendingStructuresLink = ['6093ba0e2ab5775cfc01fffe'];
      userServiceMock.create.mockResolvedValueOnce(usersMockData[0]);
      const result = await controller.create(createUserDto);

      expect(userCreateSpyer).toBeCalledTimes(1);
      expect(structureFindOneSpyer).toBeCalledTimes(1);
      expect(updateStructureLinkedClaimSpyer).toBeCalledTimes(1);
      expect(sendAdminStructureNotificationSpyer).toBeCalledTimes(1);
      expect(tempUserFindOneSpyer).toBeCalledTimes(1);
      expect(tempUserDeleteSpyer).toBeCalledTimes(0);
      expect(result).toEqual(usersMockData[0]);
    });

    it('should create user with temp user', async () => {
      const userCreateSpyer = jest.spyOn(userServiceMock, 'create');
      const structureFindOneSpyer = jest.spyOn(structureServiceMock, 'findOne');
      const updateStructureLinkedClaimSpyer = jest.spyOn(userServiceMock, 'updateStructureLinkedClaim');
      const sendAdminStructureNotificationSpyer = jest.spyOn(structureServiceMock, 'sendAdminStructureNotification');
      const tempUserFindOneSpyer = jest.spyOn(tempUserServiceMock, 'findOne');
      const tempUserDeleteSpyer = jest.spyOn(tempUserServiceMock, 'delete');

      const createUserDto = new CreateUserDto();
      createUserDto.pendingStructuresLink = [];
      userServiceMock.create.mockResolvedValueOnce(usersMockData[0]);
      tempUserServiceMock.findOne.mockResolvedValueOnce({ email: 'test@test.com', pendingStructuresLink: [] });
      const result = await controller.create(createUserDto);

      expect(userCreateSpyer).toBeCalledTimes(1);
      expect(structureFindOneSpyer).toBeCalledTimes(0);
      expect(updateStructureLinkedClaimSpyer).toBeCalledTimes(0);
      expect(sendAdminStructureNotificationSpyer).toBeCalledTimes(0);
      expect(tempUserFindOneSpyer).toBeCalledTimes(1);
      expect(tempUserDeleteSpyer).toBeCalledTimes(1);
      expect(result).toEqual(usersMockData[0]);
    });
  });

  describe('validateUser', () => {
    it('should call validateUser', async () => {
      const spyer = jest.spyOn(userServiceMock, 'validateUser');
      await controller.validateUser({ id: 1 }, 'token');
      expect(spyer).toBeCalledTimes(1);
    });
  });

  describe('changePassword', () => {
    it('should call changeUserPassword', async () => {
      const spyer = jest.spyOn(userServiceMock, 'changeUserPassword');
      await controller.changePassword({ user: { _id: '36', email: 'a@a.com' } }, new PasswordChangeDto());
      expect(spyer).toBeCalledTimes(1);
    });
  });

  describe('changeEmail', () => {
    it('should call changeUserEmail', async () => {
      const spyer = jest.spyOn(userServiceMock, 'changeUserEmail');
      await controller.changeEmail(new EmailChangeDto());
      expect(spyer).toBeCalledTimes(1);
    });
  });

  describe('verifyAndUpdateEmail', () => {
    it('should call verifyAndUpdateUserEmail', async () => {
      const spyer = jest.spyOn(userServiceMock, 'verifyAndUpdateUserEmail');
      await controller.verifyAndUpdateEmail('token');
      expect(spyer).toBeCalledTimes(1);
    });
  });

  describe('resetPassword', () => {
    it('should call sendResetPasswordEmail', async () => {
      const spyer = jest.spyOn(userServiceMock, 'sendResetPasswordEmail');
      await controller.resetPassword(new PasswordResetDto());
      expect(spyer).toBeCalledTimes(1);
    });
  });

  describe('resetPasswordApply', () => {
    it('should call validatePasswordResetToken', async () => {
      const spyer = jest.spyOn(userServiceMock, 'validatePasswordResetToken');
      await controller.resetPasswordApply(new PasswordResetApplyDto());
      expect(spyer).toBeCalledTimes(1);
    });
  });

  describe('verifyUserExist', () => {
    it('should call verifyUserExist', async () => {
      const spyer = jest.spyOn(userServiceMock, 'verifyUserExist');
      await controller.verifyUserExist({ user: { _id: '36', email: 'a@a.com' } }, { newMail: 'test@test.com' });
      expect(spyer).toBeCalledTimes(1);
    });
  });

  describe('delete', () => {
    it('should not call isStructureClaimed if no structures are linked', async () => {
      const deleteOneSpyer = jest.spyOn(userServiceMock, 'deleteOne');
      const isStructureClaimedSpyer = jest.spyOn(userServiceMock, 'isStructureClaimed');
      const userWithoutStructure = usersMockData[2];
      userServiceMock.deleteOne.mockResolvedValueOnce(userWithoutStructure);
      await controller.delete({ user: { _id: '36', email: 'a@a.com' } });
      expect(deleteOneSpyer).toBeCalledTimes(1);
      expect(isStructureClaimedSpyer).toBeCalledTimes(0);
    });

    it('should call isStructureClaimed for each structure linked', async () => {
      const userDeleteOneSpyer = jest.spyOn(userServiceMock, 'deleteOne');
      const isStructureClaimedSpyer = jest.spyOn(userServiceMock, 'isStructureClaimed');
      const structureFindOne = jest.spyOn(structureServiceMock, 'findOne');
      const userWithThreeStructures = usersMockData[3];
      userServiceMock.deleteOne.mockResolvedValueOnce(userWithThreeStructures);
      userServiceMock.isStructureClaimed.mockResolvedValue(null);
      userServiceMock.isStructureClaimed.mockResolvedValueOnce(userWithThreeStructures);
      await controller.delete({ user: { _id: '36', email: 'a@a.com' } });
      expect(userDeleteOneSpyer).toBeCalledTimes(1);
      expect(isStructureClaimedSpyer).toBeCalledTimes(3);
      expect(structureFindOne).toBeCalledTimes(2);
    });
  });

  describe('dataConsentValidation', () => {
    it('should call getAllDataConsentPendingStructures', async () => {
      const spyer = jest.spyOn(structureServiceMock, 'getAllDataConsentPendingStructures');
      await controller.dataConsentValidation({ user: { _id: '36', email: 'a@a.com' } });
      expect(spyer).toBeCalledTimes(1);
    });
  });

  describe('get user', () => {
    it('should return user', async () => {
      const spyer = jest.spyOn(userServiceMock, 'findById');
      userServiceMock.findById.mockResolvedValueOnce({ _id: '36', email: 'a@a.com' });
      const result = await controller.getUser(1);
      expect(result).toEqual({ _id: '36', email: 'a@a.com' });
      expect(spyer).toBeCalledTimes(1);
    });

    it('should throw error if no user found', async () => {
      const spyer = jest.spyOn(userServiceMock, 'findById');
      userServiceMock.findById.mockResolvedValueOnce(null);

      try {
        await controller.getUser(1);
        expect(true).toBe(false);
      } catch (error) {
        expect(error.message).toBe('User does not exist');
        expect(error.status).toBe(HttpStatus.NOT_FOUND);
      }
      expect(spyer).toBeCalledTimes(1);
    });
  });

  describe('update description', () => {
    it('should call updateDescription', async () => {
      const spyer = jest.spyOn(userServiceMock, 'updateDescription');
      await controller.updateDescription({ user: { _id: '36', email: 'a@a.com' } }, new DescriptionDto());
      expect(spyer).toBeCalledTimes(1);
    });
  });
});
