import { Injectable, Logger } from '@nestjs/common';
import { ElasticsearchService } from '@nestjs/elasticsearch';
import { es_settings_homemade_french } from '../../shared/utils';
import { UserRegistrySearchBody } from '../interfaces/userRegistry-search-body.interface';
import { UserRegistrySearchResult } from '../interfaces/userRegistry-search-response.interface';
import { IUserRegistry } from '../interfaces/userRegistry.interface';

@Injectable()
export class UserRegistrySearchService {
  private index = 'user';
  private readonly logger = new Logger(UserRegistrySearchService.name);

  constructor(private readonly elasticsearchService: ElasticsearchService) {}

  public async indexUserRegistry(user: IUserRegistry): Promise<IUserRegistry> {
    this.logger.debug(`indexUser ${user.name} ${user.surname}`);
    this.elasticsearchService.index<UserRegistrySearchResult, UserRegistrySearchBody>({
      index: this.index,
      id: user._id,
      body: {
        name: user.name,
        id: user._id,
        surname: user.surname,
        job: user.job,
        employer: user.employer,
      },
    });
    return user;
  }

  public async createUserRegistryIndex(): Promise<any> {
    this.logger.debug('createUserRegistryIndex');
    return this.elasticsearchService.indices.create({
      index: this.index,
      body: {
        mappings: {
          dynamic_templates: [
            {
              strings: {
                match_mapping_type: 'string',
                mapping: {
                  type: 'text',
                  analyzer: 'homemade_french',
                  search_analyzer: 'homemade_french_stopless',
                },
              },
            },
          ],
        },
        settings: es_settings_homemade_french,
      },
    });
  }

  public async dropIndex(): Promise<any> {
    this.logger.debug('dropIndex');
    const foundIndexes = (
      await this.elasticsearchService.indices.exists({
        index: this.index,
      })
    ).body;
    if (foundIndexes) {
      return this.elasticsearchService.indices.delete({
        index: this.index,
      });
    }
  }

  public async deleteIndex(user: IUserRegistry): Promise<IUserRegistry> {
    this.logger.debug('deleteIndex');
    this.elasticsearchService.delete<UserRegistrySearchResult, IUserRegistry>({
      index: this.index,
      id: user._id,
    });
    return user;
  }

  public async refreshIndexUserRegistry(): Promise<any> {
    this.logger.debug('refreshIndexUserRegistry');
    return this.elasticsearchService.indices.refresh({
      index: this.index,
    });
  }

  public async search(searchString: string): Promise<any[]> {
    this.logger.debug(`search user with query: ${searchString}`);
    searchString = searchString ? searchString + '*' : '*';
    const { body } = await this.elasticsearchService.search<UserRegistrySearchResult>({
      index: this.index,
      body: {
        from: 0,
        size: 200,
        query: {
          query_string: {
            analyze_wildcard: 'true',
            query: searchString,
            fields: ['name', 'surname'],
            fuzziness: 'AUTO',
          },
        },
      },
    });
    const sortedHits = body.hits.hits.filter(function (elem) {
      return elem._score >= body.hits.max_score / 1.5;
    });
    return sortedHits.map((item) => item._source);
  }

  public async update(user: IUserRegistry): Promise<any> {
    this.logger.debug(`Updates user : ${user._id} `);
    return this.elasticsearchService.update({
      index: this.index,
      id: user._id,
      body: {
        doc: {
          name: user.name,
          surname: user.surname,
          job: user.job,
          employer: user.employer,
        },
      },
    });
  }
}
