import { HttpModule } from '@nestjs/axios';
import { HttpException, HttpStatus } from '@nestjs/common';
import { getModelToken } from '@nestjs/mongoose';
import { Test, TestingModule } from '@nestjs/testing';
import * as bcrypt from 'bcrypt';
import { Types } from 'mongoose';
import { employersMockData } from '../../../test/mock/data/employers.mock.data';
import { personalOffersDataMock } from '../../../test/mock/data/personalOffers.mock.data';
import { userDetails, usersMockData } from '../../../test/mock/data/users.mock.data';
import { MailerMockService } from '../../../test/mock/services/mailer.mock.service';
import { LoginDto } from '../../auth/login-dto';
import { ConfigurationModule } from '../../configuration/configuration.module';
import { MailerModule } from '../../mailer/mailer.module';
import { MailerService } from '../../mailer/mailer.service';
import { PersonalOfferDocument } from '../../personal-offers/schemas/personal-offer.schema';
import { EmailChangeDto } from '../dto/change-email.dto';
import { CreateUserDto } from '../dto/create-user.dto';
import { DescriptionDto } from '../dto/description.dto';
import { UpdateDetailsDto } from '../dto/update-details.dto';
import { IUser } from '../interfaces/user.interface';
import { EmployerDocument } from '../schemas/employer.schema';
import { JobDocument } from '../schemas/job.schema';
import { User } from '../schemas/user.schema';
import { UserRegistrySearchService } from './userRegistry-search.service';
import { UsersService } from './users.service';

function hashPassword() {
  return bcrypt.hashSync(process.env.USER_PWD, process.env.SALT);
}

type mockUserModelFunctionTypes = {
  create: jest.Mock<any, any>;
  deleteOne: jest.Mock<any, any>;
  exec: jest.Mock<any, any>;
  find: jest.Mock<any, any>;
  findById: jest.Mock<any, any>;
  findByIdAndUpdate: jest.Mock<any, any>;
  findOne: jest.Mock<any, any>;
  findPopulatedUserRegistryById: jest.Mock<any, any>;
  limit: jest.Mock<any, any>;
  populate: jest.Mock<any, any>;
  select: jest.Mock<any, any>;
  sort: jest.Mock<any, any>;
  updateMany: jest.Mock<any, any>;
  updateOne: jest.Mock<any, any>;
};

const mockUserModel: mockUserModelFunctionTypes = {
  create: jest.fn(),
  deleteOne: jest.fn(),
  exec: jest.fn(),
  find: jest.fn(() => mockUserModel),
  findById: jest.fn(),
  findByIdAndUpdate: jest.fn(() => mockUserModel),
  findOne: jest.fn(() => mockUserModel),
  findPopulatedUserRegistryById: jest.fn(() => mockUserModel),
  limit: jest.fn(() => mockUserModel),
  populate: jest.fn(() => mockUserModel),
  select: jest.fn(() => mockUserModel),
  sort: jest.fn(() => mockUserModel),
  updateMany: jest.fn(),
  updateOne: jest.fn(),
};

const mockUserRegistrySearchService = {
  createindexUserRegistryIndex: jest.fn(),
  deleteIndex: jest.fn(),
  dropIndex: jest.fn(),
  indexUserRegistry: jest.fn(),
  isStrongPassword: jest.fn(),
  search: jest.fn(),
  update: jest.fn(),
};

const createUserDto: CreateUserDto = {
  email: 'jacques.dupont@mii.com',
  password: 'test1A!!',
  name: 'Jacques',
  surname: 'Dupont',
  phone: '06 06 06 06 06',
  structuresLink: ['61e9260c2ac971550065e262', '61e9260b2ac971550065e261'],
};

const mockUser: User = {
  role: 0,
  validationToken:
    'cf1c74c22cedb6b575945098db42d2f493fb759c9142c6aff7980f252886f36ee086574ee99a06bc99119079257116c959c8ec870949cebdef2b293666dbca42',
  emailVerified: false,
  email: 'jacques.dupont@mii.com',
  password: hashPassword(),
  newEmail: '',
  changeEmailToken: '',
  resetPasswordToken: null,
  structuresLink: [],
  structureOutdatedMailSent: [],
  pendingStructuresLink: [],
  personalOffers: [],
  name: 'Jacques',
  surname: 'Dupont',
  phone: '06 06 06 06 06',
  createdAt: new Date('2022-05-25T09:48:28.824Z'),
  employer: {
    name: 'test',
    validated: true,
  },
  job: {
    name: 'test',
    validated: true,
    hasPersonalOffer: false,
  },
  unattachedSince: null,
};

describe('UsersService', () => {
  let service: UsersService;

  beforeEach(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [HttpModule, MailerModule, ConfigurationModule],
      providers: [
        UsersService,
        UserRegistrySearchService,
        {
          provide: UserRegistrySearchService,
          useValue: mockUserRegistrySearchService,
        },
        {
          provide: getModelToken('User'),
          useValue: mockUserModel,
        },
        { provide: MailerService, useClass: MailerMockService },
      ],
    }).compile();

    service = module.get<UsersService>(UsersService);
  });
  afterEach(() => {
    jest.clearAllMocks();
  });

  it('UsersService should be defined', () => {
    expect(service).toBeDefined();
  });

  describe('create', () => {
    it('should not create a User, already exist', async () => {
      jest.spyOn(service, 'findOne').mockImplementationOnce(async (): Promise<any> => createUserDto);
      try {
        await service.create(createUserDto);
        expect(true).toBe(false);
      } catch (error) {
        expect(error.status).toBe(HttpStatus.BAD_REQUEST);
        expect(error.message).toBe('User already exists');
      }
    });

    it('should not create User, weak password', async () => {
      jest.spyOn(service, 'findOne').mockImplementationOnce(async (): Promise<any> => null);
      jest.spyOn(service, 'isStrongPassword').mockImplementationOnce(() => false);
      try {
        await service.create(createUserDto);
        expect(true).toBe(false);
      } catch (error) {
        expect(error.status).toBe(HttpStatus.UNPROCESSABLE_ENTITY);
        expect(error.message).toBe(
          'Weak password, it must contain one lowercase alphabetical character, one uppercase alphabetical character, one numeric character, one special character and be eight characters or longer'
        );
      }
    });
    it('should create a User', async () => {
      jest.spyOn(service, 'findOne').mockImplementationOnce(async (): Promise<any> => null);
      jest.spyOn(service, 'isStrongPassword').mockImplementationOnce(() => true);
      //TODO mock new userModal(createUserDto)
      return;
      jest.spyOn(service, 'findOne').mockImplementationOnce(async (): Promise<any> => mockUser);
      expect(await service.create(createUserDto)).toBe(mockUser);
    });
  });

  describe('isStrongPassword', () => {
    it('should return false with "AZERTYU1" (no lowercase character)', () => {
      expect(service.isStrongPassword('AZERTY')).toBe(false);
    });
    it('should return false with "azertyu1" (no uppercase character)', () => {
      expect(service.isStrongPassword('azerty')).toBe(false);
    });
    it('should return false with "azertyui" (no number)', () => {
      expect(service.isStrongPassword('azerty')).toBe(false);
    });
    it('should return false with "azertyu1" (no special character)', () => {
      expect(service.isStrongPassword('azertyu1')).toBe(false);
    });
    it('should return false with "azertyu1" (no special character)', () => {
      expect(service.isStrongPassword('azertyu1')).toBe(false);
    });
    it('should return false with "azer" (too short)', () => {
      expect(service.isStrongPassword('azer')).toBe(false);
    });
    it('should return true with "Azerty1!', () => {
      expect(service.isStrongPassword('Azerty1!')).toBe(true);
    });
  });

  describe('findOne', () => {
    it('should find one user', async () => {
      mockUserModel.exec.mockResolvedValue(mockUser);
      expect(await service.findOne('some@mail.com')).toBe(mockUser);
    });
  });

  describe('findAll', () => {
    it('should find all users', async () => {
      mockUserModel.exec.mockResolvedValue(usersMockData);
      expect(await service.findAll()).toBe(usersMockData);
    });
  });

  describe('findByLogin', () => {
    const loginDto: LoginDto = { email: 'jacques.dupont@mii.com', password: 'test1A!!' };
    it('should fail, no user found', async () => {
      jest.spyOn(service, 'findOne').mockResolvedValue(null);
      try {
        await service.findByLogin(loginDto);
        expect(true).toBe(false);
      } catch (error) {
        expect(error.status).toBe(HttpStatus.UNAUTHORIZED);
        expect(error.message).toBe('Invalid credentials');
      }
    });

    it('should fail, wrong password', async () => {
      jest.spyOn(service, 'findOne').mockResolvedValue(mockUser as IUser);
      //TODO mock private function comparePassword ? -> false
      return true;
      try {
        await service.findByLogin(loginDto);
        expect(true).toBe(false);
      } catch (error) {
        expect(error.status).toBe(HttpStatus.UNAUTHORIZED);
        expect(error.message).toBe('Invalid credentials');
      }
    });

    it('should find', async () => {
      mockUserModel.findOne.mockResolvedValue(mockUser);
      jest.spyOn(service, 'findOne').mockResolvedValue(mockUser as IUser);
      //TODO mock private function comparePassword ? -> true
      return true;
      expect(await service.findByLogin(loginDto)).toBe(mockUser);
    });

    it('wrong password, should be unauthorized issue', async () => {
      const result: HttpException = new HttpException('Invalid credentials', HttpStatus.UNAUTHORIZED);
      const loginDto: LoginDto = { email: 'jacques.dupont@mii.com', password: 'test1A!!!' }; //NOSONAR
      jest.spyOn(service, 'findByLogin').mockImplementation(async (): Promise<any> => result);
      expect(await service.findByLogin(loginDto)).toBe(result);
    });
  });

  describe('validateUser', () => {
    it('should not validateUser', async () => {
      const result = new HttpException('Invalid token', HttpStatus.UNAUTHORIZED);
      jest.spyOn(service, 'validateUser').mockImplementation(async (): Promise<HttpException> => result);
      expect(await service.validateUser('add3d', 'qdqdqdqd185')).toBe(result);
    });
  });

  describe('changeUserPassword', () => {
    it('should not change password', async () => {
      const result = new HttpException('Invalid token', HttpStatus.UNAUTHORIZED);
      jest.spyOn(service, 'changeUserPassword').mockImplementation(async (): Promise<HttpException> => result);
      expect(await service.changeUserPassword('add3d', 'azertyU1', 'azertyU1!d')).toBe(result);
    });
    it('should not change password', async () => {
      const result = new HttpException('Invalid token', HttpStatus.UNPROCESSABLE_ENTITY);
      jest.spyOn(service, 'changeUserPassword').mockImplementation(async (): Promise<HttpException> => result);
      expect(await service.changeUserPassword('add3d', 'azertyU1!d', 'a')).toBe(result);
    });
    it('should change password', async () => {
      const result = new HttpException('Invalid token', HttpStatus.CREATED);
      jest.spyOn(service, 'changeUserPassword').mockImplementation(async (): Promise<HttpException> => result);
      expect(await service.changeUserPassword('add3d', 'azertyU1!d', 'azertyU1!d')).toBe(result);
    });
  });

  describe('changeUserEmail', () => {
    it('should find and add token', async () => {
      const result = {
        validationToken: '',
        emailVerified: true,
        email: 'jacques.dupont@mii.com',
        password: hashPassword(),
        role: 0,
        newEmail: 'test.dupont@mail.com',
        resetPasswordToken: '',
        structuresLink: [],
        pendingStructuresLink: [],
        structureOutdatedMailSent: [],
        personalOffers: [],
        name: 'Jacques',
        surname: 'Dupont',
        phone: '06 06 06 06 06',
        createdAt: new Date('2022-05-25T09:48:28.824Z'),
        changeEmailToken:
          '9bb3542bdc5ca8801ad4cee00403c1052bc95dee768dcbb65b1f719870578ed79f71f52fdc3e6bf02fd200a72b8b6f56fc26950df30c8cd7e427a485f80181b9',
        employer: {
          name: 'test',
          validated: true,
        },
        job: {
          name: 'test',
          validated: true,
          hasPersonalOffer: false,
        },
        unattachedSince: null,
      };
      const emailDto: EmailChangeDto = { newEmail: 'test.dupont@mail.com', oldEmail: 'jacques.dupont@mii.com' }; //NOSONAR
      jest.spyOn(service, 'changeUserEmail').mockImplementation(async (): Promise<User> => result);
      expect(await service.changeUserEmail(emailDto)).toBe(result);
    });
    it('user does not exist, should be unauthorized issue', async () => {
      const result: HttpException = new HttpException('Email sent if account exist', HttpStatus.UNAUTHORIZED);
      const emailDto: EmailChangeDto = { newEmail: 'test.dupont@mail.com', oldEmail: 'jacques.dupont@mii.com' }; //NOSONAR
      jest.spyOn(service, 'changeUserEmail').mockImplementation(async (): Promise<any> => result);
      expect(await service.changeUserEmail(emailDto)).toBe(result);
    });
    it('email already used, should be not acceptable issue', async () => {
      const result: HttpException = new HttpException('Email already used', HttpStatus.NOT_ACCEPTABLE);
      const emailDto: EmailChangeDto = { newEmail: 'jacques.dupont@mii.com', oldEmail: 'jacques.dupont@mii.com' }; //NOSONAR
      jest.spyOn(service, 'changeUserEmail').mockImplementation(async (): Promise<any> => result);
      expect(await service.changeUserEmail(emailDto)).toBe(result);
    });
    it('should change email', async () => {
      const result = {
        validationToken: '',
        emailVerified: true,
        email: 'test.dupont@mail.com',
        password: hashPassword(),
        role: 0,
        newEmail: '',
        resetPasswordToken: '',
        changeEmailToken: '',
        structuresLink: [],
        pendingStructuresLink: [],
        structureOutdatedMailSent: [],
        personalOffers: [],
        name: 'Jacques',
        surname: 'Dupont',
        phone: '06 06 06 06 06',
        createdAt: new Date('2022-05-25T09:48:28.824Z'),
        employer: {
          name: 'test',
          validated: true,
        },
        job: {
          name: 'test',
          validated: true,
          hasPersonalOffer: false,
        },
        unattachedSince: null,
      };
      const token =
        '9bb3542bdc5ca8801ad4cee00403c1052bc95dee768dcbb65b1f719870578ed79f71f52fdc3e6bf02fd200a72b8b6f56fc26950df30c8cd7e427a485f80181b9'; //NOSONAR
      jest.spyOn(service, 'verifyAndUpdateUserEmail').mockImplementation(async (): Promise<User> => result);
      expect(await service.verifyAndUpdateUserEmail(token)).toBe(result);
    });
    it('should not change email', async () => {
      const result: HttpException = new HttpException('Invalid token', HttpStatus.UNAUTHORIZED);
      const token = '9bb3542bdc5ca8801aa72b8b6f56fc26950df30c8cd7e427a485f80181b9FAKETOKEN'; //NOSONAR
      jest.spyOn(service, 'verifyAndUpdateUserEmail').mockImplementation(async (): Promise<any> => result);
      expect(await service.verifyAndUpdateUserEmail(token)).toBe(result);
    });
  });

  describe('sendResetPasswordEmail', () => {
    it('should not send email', async () => {
      const result = new HttpException('Email sent if account exist', HttpStatus.OK);
      jest.spyOn(service, 'sendResetPasswordEmail').mockImplementation(async (): Promise<HttpException> => result);
      expect(await service.sendResetPasswordEmail('test@mii.com')).toBe(result);
    });

    it('should send email', async () => {
      const result = new HttpException('Email sent if account exist', HttpStatus.OK);
      jest.spyOn(service, 'sendResetPasswordEmail').mockImplementation(async (): Promise<HttpException> => result);
      expect(await service.sendResetPasswordEmail('test@mii.com')).toBe(result);
    });
  });

  describe('validatePasswordResetToken', () => {
    it('should not validate new password: token does`nt exist', async () => {
      const result = new HttpException('Invalid token', HttpStatus.UNAUTHORIZED);
      jest.spyOn(service, 'validatePasswordResetToken').mockImplementation(async (): Promise<HttpException> => result);
      expect(
        await service.validatePasswordResetToken(
          'test@mii.com',
          '5def4cb41106f89c212679e164911776618bd529e4f78e2883f7dd01776612a1b4a2ad7edabf2a3e3638aa605966c7a4b69d5f07d9617334e58332ba5f9305'
        )
      ).toBe(result);
    });

    it('should not validate new password: weak password', async () => {
      const result = new HttpException(
        'Weak password, it must contain one lowercase alphabetical character, one uppercase alphabetical character, one numeric character, one special character and be eight characters or longer',
        HttpStatus.UNPROCESSABLE_ENTITY
      );
      jest.spyOn(service, 'validatePasswordResetToken').mockImplementation(async (): Promise<HttpException> => result);
      expect(
        await service.validatePasswordResetToken(
          'test@mii.com',
          '5def4cb41106f89c212679e164911776618bd529e4f78e2883f7dd01776612a1b4a2ad7edabf2a3e3638aa605966c7a4b69d5f07d9617334e58332ba5f9305a6'
        )
      ).toBe(result);
    });

    it('should validate new password', async () => {
      const result = new HttpException('Password Reset', HttpStatus.OK);
      jest.spyOn(service, 'validatePasswordResetToken').mockImplementation(async (): Promise<HttpException> => result);
      expect(
        await service.validatePasswordResetToken(
          'test@mii.com',
          '5def4cb41106f89c212679e164911776618bd529e4f78e2883f7dd01776612a1b4a2ad7edabf2a3e3638aa605966c7a4b69d5f07d9617334e58332ba5f9305a6'
        )
      ).toBe(result);
    });

    it('should return structureLink tab ', async () => {
      const result = [53];
      jest.spyOn(service, 'updateStructureLinked').mockImplementation(async (): Promise<any> => result);
      expect(await service.updateStructureLinked('test@mii.com', '6001a37716b08100062e4160')).toBe(result);
    });

    it('should return invalid User ', async () => {
      const result = new HttpException('Invalid user', HttpStatus.NOT_FOUND);
      jest.spyOn(service, 'updateStructureLinked').mockImplementation(async (): Promise<any> => result);
      expect(await service.updateStructureLinked('test@mii.com', '6001a37716b08100062e4160')).toBe(result);
    });
  });

  it('should find All Unattacheduser', async () => {
    return;
  });

  it('should find attached users', async () => {
    return;
  });

  it('should find UnVerified Users', async () => {
    return;
  });

  describe('addPersonalOffer', () => {
    const personalOfferDocumentMock: PersonalOfferDocument = personalOffersDataMock[0] as PersonalOfferDocument;
    it('should add personal offer to the user', async () => {
      jest.spyOn(service, 'findById').mockResolvedValue(usersMockData[0]);
      const expectedResult = { ...usersMockData[0], personalOffers: [personalOfferDocumentMock] };
      expect(await service.addPersonalOffer('6036721022462b001334c4bb', personalOfferDocumentMock)).toEqual(
        expectedResult
      );
    });
    it('should return exception if user is not found for given id', async () => {
      jest.spyOn(service, 'findById').mockResolvedValue(null);
      try {
        await service.addPersonalOffer('abcd', personalOfferDocumentMock);
        // Fail test if above expression doesn't throw anything.
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toBe('User not found for the personal offer attachment');
        expect(e.status).toBe(400);
      }
    });
    it('should return exception if personal offer already exists in the user', async () => {
      jest.spyOn(service, 'findById').mockResolvedValue(usersMockData[1]);
      try {
        await service.addPersonalOffer('6093ba0e2ab5775cfc01ed3e', personalOfferDocumentMock);
        // Fail test if above expression doesn't throw anything.
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toBe('Personal offer already exist in the user');
        expect(e.status).toBe(400);
      }
    });
  });
  describe('getStructureOwnersDetails', () => {
    const userDetailsMock: IUser[] = userDetails;
    it('should get Structure Owners Details', async () => {
      jest.spyOn(service, 'getStructureOwnersDetails').mockResolvedValue(userDetails);
      expect(await service.getStructureOwnersDetails('627b85aea0466f0f132e1599', null)).toEqual(userDetailsMock);
    });
  });

  describe('user employer', () => {
    it('should replace employer with a new one', async () => {
      const spyer = jest.spyOn(mockUserModel, 'updateMany');

      mockUserModel.updateMany.mockReturnThis();
      mockUserModel.exec.mockResolvedValueOnce([usersMockData[0]]);
      await service.replaceEmployers(employersMockData[0], employersMockData[1]);
      expect(spyer).toBeCalledTimes(1);
    });

    it('should return true if a user is linked to a given employer', async () => {
      return;
    });
    it('should fetch users attached to a given employer', async () => {
      return;
    });
    it('should populate an employer with a list of attached users', async () => {
      return;
    });
    it("should update user's employer ", async () => {
      return;
    });
  });

  describe('user job', () => {
    it('should replace job with a new one', async () => {
      return;
    });

    it('should return true if a user is linked to a given job', async () => {
      return;
    });
    it('should fetch users attached to a given job', async () => {
      return;
    });
    it('should populate an job with a list of attached users', async () => {
      return;
    });
    it("should update user's job ", async () => {
      return;
    });
  });

  describe('updateUserProfile', () => {
    const employer = {
      _id: Types.ObjectId('6231aefe76598527c8d0b5ba'),
      name: 'Metro',
      validated: true,
    } as EmployerDocument;
    const job = {
      _id: Types.ObjectId('6231aefe76598527c8d0b5ba'),
      name: 'Developer',
      validated: true,
    } as JobDocument;

    it('should updateUserProfile', async () => {
      const spyer = jest.spyOn(mockUserRegistrySearchService, 'update');
      mockUserModel.findOne.mockReturnThis();
      mockUserModel.updateOne.mockResolvedValueOnce(usersMockData[0]);
      const result = await service.updateUserProfile(Types.ObjectId('627b85aea0466f0f132e1599'), employer, job);
      expect(spyer).toBeCalledTimes(1);
      expect(result).toEqual(usersMockData[0]);
    });
    it('should not updateUserProfile', async () => {
      const spyer = jest.spyOn(mockUserRegistrySearchService, 'update');
      mockUserModel.updateOne.mockResolvedValueOnce(null);
      const result = await service.updateUserProfile(Types.ObjectId('627b85aea0466f0f132e1599'), employer, job);
      expect(spyer).toBeCalledTimes(0);
      expect(result).toEqual(null);
    });
  });

  describe('updateUserJob', () => {
    const job = {
      _id: Types.ObjectId('6231aefe76598527c8d0b5ba'),
      name: 'Developer',
      validated: true,
    } as JobDocument;
    it('should updateUserJob', async () => {
      const spyer = jest.spyOn(mockUserRegistrySearchService, 'update');
      mockUserModel.updateOne.mockResolvedValueOnce(usersMockData[0]);
      mockUserModel.findOne.mockReturnThis();
      const result = await service.updateUserJob(Types.ObjectId('627b85aea0466f0f132e1599'), job);
      expect(spyer).toBeCalledTimes(1);
      expect(result).toEqual(usersMockData[0]);
    });
    it('should not updateUserJob', async () => {
      const spyer = jest.spyOn(mockUserRegistrySearchService, 'update');
      mockUserModel.updateOne.mockResolvedValueOnce(null);
      const result = await service.updateUserJob(Types.ObjectId('627b85aea0466f0f132e1599'), job);
      expect(spyer).toBeCalledTimes(0);
      expect(result).toEqual(null);
    });
  });

  describe('updateUserEmployer', () => {
    const employer = {
      _id: Types.ObjectId('6231aefe76598527c8d0b5ba'),
      name: 'Metro',
      validated: true,
    } as EmployerDocument;
    it('should updateUserEmployer', async () => {
      const spyer = jest.spyOn(mockUserRegistrySearchService, 'update');
      mockUserModel.updateOne.mockResolvedValueOnce(usersMockData[0]);
      const result = await service.updateUserEmployer(Types.ObjectId('627b85aea0466f0f132e1599'), employer);
      expect(spyer).toBeCalledTimes(1);
      expect(result).toEqual(usersMockData[0]);
    });
    it('should not updateUserEmployer', async () => {
      const spyer = jest.spyOn(mockUserRegistrySearchService, 'update');
      mockUserModel.updateOne.mockResolvedValueOnce(null);
      const result = await service.updateUserEmployer(Types.ObjectId('627b85aea0466f0f132e1599'), employer);
      expect(spyer).toBeCalledTimes(0);
      expect(result).toEqual(null);
    });
  });

  describe('updateDetails', () => {
    it('should update user details', async () => {
      const user = usersMockData[0];
      const detailsDto: UpdateDetailsDto = { name: 'Michel', surname: 'Chelmi', phone: '0601020304' };
      mockUserModel.findByIdAndUpdate.mockResolvedValueOnce({ ...user, ...detailsDto });
      mockUserModel.findPopulatedUserRegistryById.mockResolvedValueOnce({ ...user, ...detailsDto });
      mockUserModel.findOne.mockReturnThis();
      const updatedUser = await service.updateUserDetails('', { name: '', surname: '', phone: '' });
      expect(updatedUser.name).toBe(detailsDto.name);
      expect(updatedUser.surname).toBe(detailsDto.surname);
      expect(updatedUser.phone).toBe(detailsDto.phone);
    });
    it('should not found a user', async () => {
      mockUserModel.findByIdAndUpdate.mockResolvedValueOnce(null);
      try {
        await service.updateUserDetails('', { name: '', surname: '', phone: '' });
        expect(true).toBe(false);
      } catch (error) {
        expect(error.message).toBe('User not found');
        expect(error.status).toBe(HttpStatus.BAD_REQUEST);
      }
    });
  });

  describe('updateDescription', () => {
    it('should update user description', async () => {
      const user = usersMockData[0];
      const updatedDescription: DescriptionDto = { description: 'hello friend' };
      mockUserModel.findByIdAndUpdate.mockResolvedValueOnce({ ...user, ...updatedDescription });
      const updatedUser = await service.updateDescription('', { description: '' });
      expect(updatedUser.description).toBe(updatedDescription.description);
    });
    it('should not found a user', async () => {
      mockUserModel.findByIdAndUpdate.mockResolvedValueOnce(null);
      try {
        await service.updateDescription('', { description: '' });
        expect(true).toBe(false);
      } catch (error) {
        expect(error.message).toBe('User not found');
        expect(error.status).toBe(HttpStatus.BAD_REQUEST);
      }
    });
  });
});
